import os
from typing import Any
import logging

import numpy as np
from psychopy import core, event, visual

from bcipy.tasks.exceptions import InsufficientDataException

log = logging.getLogger(__name__)

SPACE_CHAR = '_'
BACKSPACE_CHAR = '<'
DEFAULT_CHANNEL_MAP = (1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1,
                       1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 1, 1, 1, 0)


def fake_copy_phrase_decision(copy_phrase, target_letter, text_task):
    """Fake Copy Phrase Decision.

    Parameters
    ----------
        copy_phrase(str): phrase to be copied
        target_letter(str): the letter supposed to be typed
        text_task(str): phrase spelled at this time

    Returns
    -------
        (next_target_letter, text_task, run) tuple
    """
    if text_task == '*':
        length_of_spelled_letters = 0
    else:
        length_of_spelled_letters = len(text_task)

    length_of_phrase = len(copy_phrase)

    if length_of_spelled_letters is 0:
        text_task = copy_phrase[length_of_spelled_letters]
    else:
        text_task += copy_phrase[length_of_spelled_letters]

    length_of_spelled_letters += 1

    # If there is still text to be spelled, update the text_task
    # and target letter
    if length_of_spelled_letters < length_of_phrase:
        next_target_letter = copy_phrase[length_of_spelled_letters]

        run = True

    # else, end the run
    else:
        run = False
        next_target_letter = None
        text_task = copy_phrase

    return next_target_letter, text_task, run


def calculate_stimulation_freq(flash_time: float) -> float:
    """Calculate Stimulation Frequency.

    In an RSVP paradigm, the sequence itself will produce an
        SSVEP response to the stimulation. Here we calculate
        what that frequency should be based in the presentation
        time.

    PARAMETERS
    ----------
    :param: flash_time: time in seconds to present RSVP sequence letters
    :returns: frequency: stimulation frequency of the sequence
    """

    # We want to know how many stimuli will present in a second
    return 1 / flash_time


def alphabet(parameters=None, include_path=True):
    """Alphabet.

    Function used to standardize the symbols we use as alphabet.

    Returns
    -------
        array of letters.
    """
    if parameters:
        if not parameters['is_txt_stim']:
            # construct an array of paths to images
            path = parameters['path_to_presentation_images']
            stimulus_array = []
            for stimulus_filename in sorted(os.listdir(path)):
                # PLUS.png is reserved for the fixation symbol
                if stimulus_filename.endswith(
                        '.png') and not stimulus_filename.endswith('PLUS.png'):
                    if include_path:
                        img = os.path.join(path, stimulus_filename)
                    else:
                        img = os.path.splitext(stimulus_filename)[0]
                    stimulus_array.append(img)
            return stimulus_array

    return ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
            'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
            BACKSPACE_CHAR, SPACE_CHAR]


def process_data_for_decision(
        sequence_timing,
        daq,
        window,
        parameters,
        first_session_stim_time,
        static_offset=None,
        buf_length=None):
    """Process Data for Decision.

    Processes the raw data (triggers and EEG) into a form that can be passed to
    signal processing and classifiers.

    Parameters
    ----------
        sequence_timing(array): array of tuples containing stimulus timing and
            text
        daq (object): data acquisition object
        window: window to reactivate if deactivated by windows
        parameters: parameters dictionary
        first_session_stim_time (float): time that the first stimuli was presented
            for the session. Used to calculate offsets.

    Returns
    -------
        (raw_data, triggers, target_info) tuple
    """

    # Get timing of the first and last stimuli
    _, first_stim_time = sequence_timing[0]
    _, last_stim_time = sequence_timing[-1]

    static_offset = static_offset or parameters['static_trigger_offset']
    # if there is an argument supplied for buffer length use that
    if buf_length:
        buffer_length = buf_length
    else:
        buffer_length = parameters['len_data_sequence_buffer']

    # get any offset calculated from the daq
    daq_offset = daq.offset

    if daq_offset:
        offset = daq_offset - first_session_stim_time + static_offset
        time1 = (first_stim_time + offset) * daq.device_info.fs
        time2 = (last_stim_time + offset + buffer_length) * daq.device_info.fs
    else:
        time1 = (first_stim_time + static_offset) * daq.device_info.fs
        time2 = (last_stim_time + static_offset + buffer_length) * daq.device_info.fs

    # Construct triggers to send off for processing
    triggers = [(text, ((timing) - first_stim_time))
                for text, timing in sequence_timing]

    # Assign labels for triggers
    # TODO: This doesn't seem useful and is misleading
    target_info = ['nontarget'] * len(triggers)

    # Define the amount of data required for any processing to occur.
    data_limit = (last_stim_time - first_stim_time + buffer_length) * daq.device_info.fs

    # Query for raw data
    try:
        # Call get_data method on daq with start/end
        raw_data = daq.get_data(start=time1, end=time2, win=window)

        # If not enough raw_data returned in the first query, let's try again
        #  using only the start param. This is known issue on Windows.
        #  #windowsbug
        if len(raw_data) < data_limit:

            # Call get_data method on daq with just start
            raw_data = daq.get_data(start=time1, win=window)

            # If there is still insufficient data returned, throw an error
            if len(raw_data) < data_limit:
                message = f'Process Data Error: Not enough data received to process. ' \
                          f'Data Limit = {data_limit}. Data received = {len(raw_data)}'
                log.error(message)
                raise InsufficientDataException(message)

        # Take only the sensor data from raw data and transpose it
        raw_data = np.array([np.array([_float_val(col) for col in record.data])
                             for record in raw_data],
                            dtype=np.float64).transpose()

    except Exception as e:
        log.error(f'Uncaught Error in Process Data for Decision: {e}')
        raise e

    return raw_data, triggers, target_info


def _float_val(col: Any) -> float:
    """Convert marker data to float values so we can put them in a
    typed np.array. The marker column has type float if it has a 0.0
    value, and would only have type str for a marker value."""
    if isinstance(col, str):
        return 1.0
    return float(col)


def trial_complete_message(win, parameters):
    """Trial Complete Message.

    Function return a TextStim Object (see Psycopy) to complete the trial.

    Parameters
    ----------

        win (object): Psychopy Window Object, should be the same as the one
            used in the experiment
        parameters (dict): Dictionary of session parameters

    Returns
    -------
        array of message_stim (trial complete message to be displayed).
    """
    message_stim = visual.TextStim(
        win=win,
        height=parameters['task_height'],
        text=parameters['trial_complete_message'],
        font=parameters['task_font'],
        pos=(float(parameters['text_pos_x']),
             float(parameters['text_pos_y'])),
        wrapWidth=None,
        color=parameters['trial_complete_message_color'],
        colorSpace='rgb',
        opacity=1, depth=-6.0)
    return [message_stim]


def print_message(window: visual.Window, message: str = "Initializing..."):
    """Draws a message on the display window using default config.

    Parameters
    ----------
        window (object): Psychopy Window Object, should be the same as the one
            used in the experiment
        parameters (dict): Dictionary of session parameters

    Returns
    -------
        TextStim object
    """
    message_stim = visual.TextStim(win=window, text=message)
    message_stim.draw()
    window.flip()
    return message_stim


def get_user_input(window, message, color, first_run=False):
    """Get User Input.

    Function returns whether or not to stop a trial. If a key of interest is
        passed (e.g. space), it will act on it.

    Parameters
    ----------

        window[psychopy task window]: task window.  *assumes wait_screen method

    Returns
    -------
        True/False: whether or not to stop a trial (based on escape key).
    """
    if not first_run:
        pause = False
        # check user input to make sure we should be going
        keys = event.getKeys(keyList=['space', 'escape'])

        if keys:
            # pause?
            if keys[0] == 'space':
                pause = True

            # escape?
            if keys[0] == 'escape':
                return False

    else:
        pause = True

    while pause:
        window.wait_screen(message, color)
        keys = event.getKeys(keyList=['space', 'escape'])

        if keys:
            if keys[0] == 'escape':
                return False
            pause = False

    return True


def trial_reshaper(trial_target_info: list,
                   timing_info: list,
                   eeg_data: np.array,
                   fs: int, k: int, mode: str,
                   offset: float = 0,
                   channel_map: tuple = DEFAULT_CHANNEL_MAP,
                   trial_length: float = 0.5) -> tuple:
    """Trial Reshaper.

    Trail reshaper is used to reshape trials, based on trial target info (target, non-target),
        timing information (sec), eeg_data, sampling rate (fs), down-sampling rate (k),
        mode of operation (ex. 'calibration'),
        offset (any calculated or hypothesized offsets in timings),
        channel map (which channels to include in reshaping) and
        trial length (length of reshaped trials in secs)

    In all modes > calibration, we assume the timing info is given in samples not seconds.

    :return (reshaped_trials, labels, num_of_sequences, trials_per_seq)

        reshaped_trials is a 3 dimensional np array where the first dimension
            is the channel, second dimension the trial/letter, and third
            dimension is time samples. So at 300hz and a presentation time of
            0.2 seconds per letter, the third dimension would have about 60
            samples.
            If this were a json structure it might look like:
                {
                    'ch1': {
                        'B' : [-647.123, -93.124, ...],
                        'T' : [...],
                        ...
                    },
                    'ch2': {
                        'B' : [39.60, -224.124, ...],
                        'T' : [...],
                        ...
                    },
                    ...
                }
    """
    try:
        # Remove the channels that we are not interested in
        channel_indexes_to_remove = []
        for channel_index in range(len(eeg_data)):
            if channel_map[channel_index] == 0:
                channel_indexes_to_remove.append(channel_index)

        eeg_data = np.delete(eeg_data,
                             channel_indexes_to_remove,
                             axis=0)
        after_filter_frequency = fs / k
        # Number of samples we are interested per trial
        num_samples = int(trial_length * after_filter_frequency)

        # MODE CALIBRATION
        if mode == 'calibration':
            trials_per_seq = []
            count = 0
            # Count every sequences trials
            for symbol_info in trial_target_info:
                if symbol_info == 'first_pres_target':
                    trials_per_seq.append(count)
                    count = 0
                elif symbol_info == 'nontarget' or 'target':
                    count += 1
                else:
                    raise Exception('Incorrectly formatted trigger file. \
                    See trial_reshaper documentation for expected input.')

            # The first element is garbage. Get rid of it.
            trials_per_seq = trials_per_seq[1:]
            # Append the last sequences trial number
            trials_per_seq.append(count)
            # Make the list a numpy array.
            trials_per_seq = np.array(trials_per_seq)

            # Mark every element in timing_info if 'first_pres_target'
            for symbol_info_index in range(len(trial_target_info)):
                if trial_target_info[symbol_info_index] == 'first_pres_target':
                    timing_info[symbol_info_index] = -1

            # Get rid of 'first_pres_target' trials information
            trial_target_info = list(filter(lambda x: x != 'first_pres_target',
                                            trial_target_info))
            timing_info = list(filter(lambda x: x != -1, timing_info))

            # triggers in seconds are mapped to triggers in number of samples.
            triggers = list(
                map(lambda x: int((x + offset) * after_filter_frequency), timing_info))

            # 3 dimensional np array first dimension is channels
            # second dimension is trials and third dimension is time samples.
            reshaped_trials = np.zeros(
                (len(eeg_data), len(triggers), num_samples))

            # Label for every trial
            labels = np.zeros(len(triggers))

            for trial in range(len(triggers)):
                # Assign targetness to labels for each trial
                if trial_target_info[trial] == 'target':
                    labels[trial] = 1

                # For every channel append filtered channel data to trials
                for channel in range(len(eeg_data)):
                    reshaped_trials[channel][trial] = \
                        eeg_data[channel][triggers[trial]:triggers[trial] + num_samples]

            num_of_sequences = int(sum(labels))

        # MODE COPY PHRASE
        elif mode == 'copy_phrase':

            # triggers in samples are mapped to triggers in number of filtered samples.
            triggers = list(map(lambda x: int((x + offset) * after_filter_frequency), timing_info))

            # 3 dimensional np array first dimension is channels
            # second dimension is trials and third dimension is time samples.
            reshaped_trials = np.zeros(
                (len(eeg_data), len(triggers), num_samples))

            # Label for every trial
            labels = np.zeros(len(triggers))

            for trial in range(len(triggers)):
                # Assign targetness to labels for each trial
                if trial_target_info[trial] == 'target':
                    labels[trial] = 1

                # For every channel append filtered channel data to trials
                for channel in range(len(eeg_data)):
                    reshaped_trials[channel][trial] = \
                        eeg_data[channel][
                        triggers[trial]:triggers[trial] + num_samples]

            # In copy phrase, num of sequence is assumed to be 1.
            num_of_sequences = 1
            # Since there is only one sequence, all trials are in the sequence
            trials_per_seq = len(triggers)

        # MODE FREE SPELL
        elif mode == 'free_spell':

            # triggers in sample are mapped to triggers in number of filtered samples.
            triggers = list(map(lambda x: int((x + offset) * after_filter_frequency), timing_info))

            # 3 dimensional np array first dimension is channels
            # second dimension is trials and third dimension is time samples.
            reshaped_trials = np.zeros(
                (len(eeg_data), len(triggers), num_samples))

            labels = None

            for trial in range(len(triggers)):

                # For every channel append filtered channel data to trials
                for channel in range(len(eeg_data)):
                    reshaped_trials[channel][trial] = \
                        eeg_data[channel][
                        triggers[trial]:triggers[trial] + num_samples]

            # In copy phrase, num of sequence is assumed to be 1.
            num_of_sequences = 1
            # Since there is only one sequence, all trials are in the sequence
            trials_per_seq = len(triggers)
        else:
            raise Exception(
                'Trial_reshaper does not work in this operating mode.')

        # Return our trials, labels and some useful information about the arrays
        return reshaped_trials, labels, num_of_sequences, trials_per_seq

    except Exception as e:
        raise Exception(
            f'Could not reshape trial for mode: {mode}, {fs}, {k}. Error: {e}')


def pause_calibration(window, display, current_index: int, parameters: dict):
    """Pause calibration.

    Pauses calibration for a given number of seconds and displays a countdown
    to the user.


    PARAMETERS
    ----------
    :param: window: Currently active PsychoPy window
    :param: display: The current display
    :param: current_index: number of trials that have already taken place
    :param: trials_before_break: number of trials before break
    :param: break_len: length of the break time (in seconds)
    :param: break_message: message to display to the user during the break

    :returns: bool: break has taken place
    """
    # Check whether or not to present a break
    trials_before_break = parameters['trials_before_break']
    break_len = parameters['break_len']
    break_message = parameters['break_message']

    if (current_index != 0) and (current_index % trials_before_break) == 0:

        # present break message for break length
        for counter in range(break_len):
            time = break_len - counter
            message = f'{break_message} {time}s'
            display.update_task_state(
                text=message,
                color_list=['white'])
            display.draw_static()
            window.flip()
            core.wait(1)
        return True

    return False



"""Defines classes that can write markers to LabStreamingLayer StreamOutlet."""
import logging
from typing import Any
import pylsl

log = logging.getLogger(__name__)

class MarkerWriter():
    """Abstract base class for an object that can be used to handle stimulus
    markers.
    """

    def push_marker(self, marker: Any, lsl_time: float = None):
        """Push the given stimulus marker for processing.

        Parameters:
            marker : any object that can be converted to a str
            lsl_time : optional float indicating the time to associate with
                the marker.
        """
        raise NotImplementedError()

    def now(self) -> float:
        """Returns current time from the clock internal to the marker
        writer."""
        raise NotImplementedError()

    def cleanup(self):
        """Performs any necessary cleanup"""
        raise NotImplementedError()


class LslMarkerWriter(MarkerWriter):
    """Writes stimulus markers to a LabStreamingLayer StreamOutlet
    using pylsl. To consume this data, the client code would need to create a
    pylsl.StreamInlet. See https://github.com/sccn/labstreaminglayer/wiki.
    """

    def __init__(self):
        super(LslMarkerWriter, self).__init__()
        markers_info = pylsl.StreamInfo("BCI Stimulus Markers",
                                        "Markers", 1, 0, 'string',
                                        "bci_stim_markers")
        self.markers_outlet = pylsl.StreamOutlet(markers_info)
        self._stamp = None

    def now(self) -> float:
        """Create the timestamp for the next marker using the pylsl
        local_clock."""
        return pylsl.local_clock()

    def push_marker(self, marker: Any, lsl_time: float = None):
        """Push the given stimulus marker for processing.
        Parameters:
            marker : any object that can be converted to a str
            lsl_time : optional float indicating the time to associate with
                the marker. If omitted, the pylsl.local_clock() will be used.
        """
        stamp = lsl_time if lsl_time is not None else self.now()
        self.markers_outlet.push_sample([str(marker)], stamp)
        log.debug("Pushing marker: %s; timestamp: %s", marker, stamp)

    def cleanup(self):
        """Cleans up the StreamOutlet."""
        del self.markers_outlet


class NullMarkerWriter(MarkerWriter):
    """MarkerWriter which doesn't write anything.

    A NullMarkerWriter can be passed in to the calling object in scenarios
    where marker handling occurs indirectly (ex. through a trigger box). By
    using a NullMarkerWriter rather than a None value, the calling
    object does not have to do additional null checks and a separation
    of concerns is maintained regarding how triggers are written for different
    devices.

    See the Null Object Design Pattern:
    https://en.wikipedia.org/wiki/Null_object_pattern
    """

    def push_marker(self, marker: Any, lsl_time: float = None):
        pass

    def now(self) -> float:
        return 0.0

    def cleanup(self):
        pass



"""Defines the CopyPhraseWrapper."""
from typing import List, Tuple

import numpy as np

from bcipy.helpers.acquisition import analysis_channels
from bcipy.helpers.task import BACKSPACE_CHAR, trial_reshaper
from bcipy.signal.model.inference import inference
from bcipy.signal.process.filter import bandpass, notch, downsample
from bcipy.tasks.rsvp.main_frame import EvidenceFusion, DecisionMaker
from bcipy.helpers.language_model import norm_domain, sym_appended, \
 equally_probable


class CopyPhraseWrapper:
    """Basic copy phrase task duty cycle wrapper.

    Given the phrases once operate() is called performs the task.
    Attr:
        min_num_seq: The minimum number of sequences to be displayed
        max_num_seq: The maximum number of sequences to be displayed
        model(pipeline): model trained using a calibration session of the
            same user.
        fs(int): sampling frequency
        k(int): down sampling rate
        alp(list[str]): symbol set of the task
        task_list(list[tuple(str,str)]): list[(phrases, initial_states)] for
            the copy phrase task
        is_txt_stim: Whether or not the stimuli are text objects
        conjugator(EvidenceFusion): fuses evidences in the task
        decision_maker(DecisionMaker): mastermind of the task
        mode(str): mode of thet task (should be copy phrase)
        d(binary): decision flag
        sti(list(tuple)): stimuli for the display
        decision_threshold: Minimum likelihood value required for a decision
        backspace_prob(float): default language model probability for the
            backspace character.
        backspace_always_shown(bool): whether or not the backspace should
            always be presented.
    """

    def __init__(self, min_num_seq, max_num_seq, signal_model=None, fs=300, k=2,
                 alp=None, evidence_names=['LM', 'ERP'],
                 task_list=[('I_LOVE_COOKIES', 'I_LOVE_')], lmodel=None,
                 is_txt_stim=True, device_name='LSL', device_channels=None,
                 stimuli_timing=[1, .2],
                 decision_threshold=0.8,
                 backspace_prob=0.05,
                 backspace_always_shown=False,
                 filter_high=45,
                 filter_low=2,
                 filter_order=2,
                 notch_filter_frequency=60):

        self.conjugator = EvidenceFusion(evidence_names, len_dist=len(alp))

        seq_constants = []
        if backspace_always_shown and BACKSPACE_CHAR in alp:
            seq_constants.append(BACKSPACE_CHAR)
        self.decision_maker = DecisionMaker(
            min_num_seq,
            max_num_seq,
            decision_threshold=decision_threshold,
            state=task_list[0][1],
            alphabet=alp,
            is_txt_stim=is_txt_stim,
            stimuli_timing=stimuli_timing,
            seq_constants=seq_constants)
        self.alp = alp
        # non-letter target labels include the fixation cross and calibration.
        self.nonletters = ['+', 'PLUS', 'calibration_trigger']
        self.valid_targets = set(self.alp)

        self.signal_model = signal_model
        self.sampling_rate = fs
        self.downsample_rate = k
        self.filter_high = filter_high
        self.filter_low = filter_low
        self.filter_order = filter_order
        self.notch_filter_frequency = notch_filter_frequency

        self.mode = 'copy_phrase'
        self.task_list = task_list
        self.lmodel = lmodel
        self.channel_map = analysis_channels(device_channels, device_name)
        self.backspace_prob = backspace_prob

    def evaluate_sequence(self, raw_data, triggers, target_info, window_length):
        """Once data is collected, infers meaning from the data.

        Args:
            raw_data(ndarray[float]): C x L eeg data where C is number of
                channels and L is the signal length
            triggers(list[tuple(str,float)]): triggers e.g. ('A', 1)
                as letter and flash time for the letter
            target_info(list[str]): target information about the stimuli
            window_length(int): The length of the time between stimuli presentation
        """
        letters, times, target_info = self.letter_info(triggers, target_info)

        # Remove 60hz noise with a notch filter
        notch_filter_data = notch.notch_filter(
            raw_data, self.sampling_rate, frequency_to_remove=self.notch_filter_frequency)

        # bandpass filter from 2-45hz
        filtered_data = bandpass.butter_bandpass_filter(
            notch_filter_data,
            self.filter_low,
            self.filter_high,
            self.sampling_rate,
            order=self.filter_order)

        # downsample
        data = downsample.downsample(filtered_data, factor=self.downsample_rate)
        x, _, _, _ = trial_reshaper(target_info, times, data, fs=self.sampling_rate,
                                    k=self.downsample_rate, mode=self.mode,
                                    channel_map=self.channel_map,
                                    trial_length=window_length)

        lik_r = inference(x, letters, self.signal_model, self.alp)
        prob = self.conjugator.update_and_fuse({'ERP': lik_r})
        decision, arg = self.decision_maker.decide(prob)

        if 'stimuli' in arg:
            sti = arg['stimuli']
        else:
            sti = None

        return decision, sti

    def letter_info(self, triggers: List[Tuple[str, float]],
                    target_info: List[str]
                    ) -> Tuple[List[str], List[float], List[str]]:
        """
        Filters out non-letters and separates timings from letters.
        Parameters:
        -----------
         triggers: triggers e.g. [['A', 0.5], ...]
                as letter and flash time for the letter
         target_info: target information about the stimuli;
            ex. ['nontarget', 'nontarget', ...]
        Returns:
        --------
            (letters, times, target_info)
        """
        letters = []
        times = []
        target_types = []

        for i, (letter, stamp) in enumerate(triggers):
            if letter not in self.nonletters:
                letters.append(letter)
                times.append(stamp)
                target_types.append(target_info[i])

        # Raise an error if the stimuli includes unexpected terms
        if not set(letters).issubset(self.valid_targets):
            invalid = set(letters).difference(self.valid_targets)
            raise Exception(
                f'unexpected letters received in copy phrase: {invalid}')

        return letters, times, target_types

    def initialize_epoch(self):
        """If a decision is made initializes the next epoch."""

        try:
            # First, reset the history for this new epoch
            self.conjugator.reset_history()

            # If there is no language model specified, mock the LM prior
            # TODO: is the probability domain correct? ERP evidence is in
            # the log domain; LM by default returns negative log domain.
            if not self.lmodel:
                # mock probabilities to be equally likely for all letters.
                overrides = {BACKSPACE_CHAR: self.backspace_prob}
                prior = equally_probable(self.alp, overrides)

            # Else, let's query the lmodel for priors
            else:
                # Get the displayed state
                # TODO: for oclm this should be a list of (sym, prob)
                update = self.decision_maker.displayed_state

                # update the lmodel and get back the priors
                lm_prior = self.lmodel.state_update(update)

                # normalize to probability domain
                lm_letter_prior = norm_domain(lm_prior['letter'])

                if BACKSPACE_CHAR in self.alp:
                    # Append backspace if missing.
                    sym = (BACKSPACE_CHAR, self.backspace_prob)
                    lm_letter_prior = sym_appended(lm_letter_prior, sym)

                # convert to format needed for evidence fusion;
                # probability value only in alphabet order.
                # TODO: ensure that probabilities still add to 1.0
                prior = [prior_prob
                         for alp_letter in self.alp
                         for prior_sym, prior_prob in lm_letter_prior
                         if alp_letter == prior_sym]

            # Try fusing the lmodel evidence
            try:
                prob_dist = self.conjugator.update_and_fuse(
                    {'LM': np.array(prior)})
            except Exception as lm_exception:
                print("Error updating language model!")
                raise lm_exception

            # Get decision maker to give us back some decisions and stimuli
            is_accepted, arg = self.decision_maker.decide(prob_dist)
            sti = arg['stimuli']

        except Exception as init_exception:
            print("Error in initialize_epoch: %s" % (init_exception))
            raise init_exception

        return is_accepted, sti



import csv
import datetime
import glob
import random

from itertools import repeat
from os import path
from os.path import basename, dirname, splitext
from typing import Dict, List, Sequence, Tuple

from psychopy import core

from bcipy.helpers.system_utils import auto_str
from bcipy.display.rsvp.mode.icon_to_icon import IconToIconDisplay
from bcipy.feedback.visual.visual_feedback import VisualFeedback, FeedbackType
from bcipy.helpers.task import (
    alphabet, fake_copy_phrase_decision, get_user_input,
    process_data_for_decision, trial_complete_message)
from bcipy.helpers.save import _save_session_related_data
from bcipy.helpers.signal_model import CopyPhraseWrapper
from bcipy.helpers.stimuli import generate_icon_match_images
from bcipy.helpers.triggers import write_triggers_from_sequence_icon_to_icon
from bcipy.tasks.task import Task
from bcipy.helpers.system_utils import auto_str


class RSVPIconToIconTask(Task):
    """RSVP Icon to Icon Matching Task.

    Parameters
    ----------
        parameters : dict,
            configuration details regarding the experiment. See parameters.json
        daq : object,
            data acquisition object initialized for the desired protocol
        file_save : str,
            path location of where to save data from the session
        signal_model : loaded pickle file,
            trained signal_model, loaded before session started
        fake : boolean, optional,
            boolean to indicate whether this is a fake session or not.
        is_word: boolean,
            boolean to indicate whether or not this is an icon to word matching session
        auc_filename: str,
            name of the loaded pickle file
    Returns
    -------
        file_save : str,
            path location of where to save data from the session
    """
    TASK_NAME = 'RSVP Icon to Icon Task'

    def __init__(self, win, daq, parameters, file_save, signal_model,
                 language_model, fake, is_word, auc_filename):
        super(RSVPIconToIconTask, self).__init__()
        self.window = win
        self.frame_rate = self.window.getActualFrameRate()
        self.parameters = parameters
        self.daq = daq
        self.static_clock = core.StaticPeriod(screenHz=self.frame_rate)
        self.experiment_clock = core.Clock()
        self.buffer_val = parameters['task_buffer_len']

        self.image_path = parameters['path_to_presentation_images']
        # Alphabet is comprised of the image base names
        self.alp = alphabet(parameters, include_path=False)

        self.rsvp = _init_icon_to_icon_display_task(
            self.parameters, self.window, self.daq, self.static_clock,
            self.experiment_clock, is_word)
        self.file_save = file_save
        self.is_word = is_word

        trigger_save_location = f'{self.file_save}/{parameters["trigger_file_name"]}'
        self.trigger_file = open(trigger_save_location, 'w+')
        self.session_save_location = f'{self.file_save}/{parameters["session_file_name"]}'

        self.wait_screen_message = parameters['wait_screen_message']
        self.wait_screen_message_color = parameters[
            'wait_screen_message_color']

        self.stim_number = parameters['stim_number']
        self.stim_length = parameters['stim_length']
        self.timing = [
            parameters['time_target'], parameters['time_cross'],
            parameters['time_flash']
        ]

        self.color = [
            parameters['target_letter_color'], parameters['fixation_color'],
            parameters['stim_color']
        ]

        self.task_info_color = parameters['task_color']

        self.stimuli_height = parameters['stim_height']

        self.eeg_buffer = parameters['eeg_buffer_len']

        self.max_seconds = parameters['max_minutes'] * 60  # convert to seconds
        self.max_seq_length = parameters['max_seq_len']
        self.max_seq_per_trial = parameters['max_seq_per_trial']
        self.fake = fake
        self.language_model = language_model
        self.signal_model = signal_model
        self.auc_filename = auc_filename

        self.task_height = parameters['task_height']

        self.is_txt_stim = False

        self.min_num_seq = parameters['min_seq_len']
        self.word_matching_text_size = parameters['word_matching_text_size']
        self.collection_window_len = parameters[
            'collection_window_after_trial_length']

        self.data_save_path = parameters['data_save_loc']

        match_type = 'Word' if self.is_word else 'Icon'
        self.session_description = f'Icon to {match_type} Matching'

    def img_path(self, alphabet_item):
        """Return the full image path for the given alphabet item. If the item
        ends with .png it's returned as is."""
        if alphabet_item.endswith('.png'):
            return alphabet_item
        return self.image_path + alphabet_item + '.png'

    def init_session_data(self, save: bool = True):
        """Initializes the session data; saved to session.json file.
        Parameters:
        -----------
            save - whether to save the data to disk.
        Returns:
        --------
            data - map with initial session data.
        """
        data = {
            'session': self.file_save,
            'session_type': self.session_description,
            'paradigm': 'RSVP',
            'epochs': {},
            'total_time_spent': self.experiment_clock.getTime(),
            'total_number_epochs': 0,
        }

        if save:
            _save_session_related_data(self.session_save_location, data)
        return data

    def await_start(self) -> bool:
        """Wait for user input to either exit or start"""
        self.logger.debug('Awaiting user start.')
        should_continue = get_user_input(
            self.rsvp,
            self.wait_screen_message,
            self.wait_screen_message_color,
            first_run=True)
        return should_continue

    def user_wants_to_continue(self) -> bool:
        """Check if user wants to continue or terminate.
        Returns True to continue."""
        should_continue = get_user_input(
            self.rsvp,
            self.wait_screen_message,
            self.wait_screen_message_color,
            first_run=False)
        if not should_continue:
            self.logger.debug('User wants to exit.')
        return should_continue

    def present_sequence(self,
                         seq: List[str],
                         durations: List[float],
                         show_target: bool = False):
        """Present the given sequence and return the trigger timing info.
        Parameters:
        ----------
          seq - list of alphabet items to present
          duration - list of durations (float) to present each item
          show_target - optional item to highlight the first sequence item as
            the target (displayed in the header and outlined).
        Returns:
        --------
            sequence timings - list of tuples representing the letter and time
                that it was presented.
        """
        # Sequences passed to rsvp to display should be a list of image paths
        # except for a word target.
        word_target = show_target and self.is_word
        self.rsvp.stimuli_sequence = [
            item if i == 0 and word_target else self.img_path(item)
            for i, item in enumerate(seq)
        ]
        self.rsvp.stimuli_timing = durations

        if self.is_word:
            if show_target:
                # Change size of target word if we are in word matching mode
                # and the word is presented.
                word_height = self.word_matching_text_size
                self.rsvp.size_list_sti = [
                    word_height if i == 0 else self.stimuli_height
                    for i, _ in enumerate(seq)
                ]
            else:
                self.rsvp.size_list_sti = []

        core.wait(self.buffer_val)

        # Present the Target and place it in the header
        if show_target:
            self.rsvp.update_task_state(self.rsvp.stimuli_sequence[0],
                                        self.task_height, 'yellow',
                                        self.rsvp.window.size, self.is_word)

        # Note that returned triggers use image basenames only.
        sequence_timing = self.rsvp.do_sequence()
        return sequence_timing

    def display_feedback(self, selection: str, correct: bool):
        """Display feedback for the given selection."""
        feedback = VisualFeedback(
            display=self.window,
            parameters=self.parameters,
            clock=self.experiment_clock)
        feedback.message_color = 'green' if correct else 'red'
        feedback.administer(
            self.img_path(selection),
            compare_assertion=None,
            message='Decision: ',
            stimuli_type=FeedbackType.IMAGE)

    def write_data(self, correct_trials: int, selections: int):
        """Write trial data to icon_data.csv in file save location."""

        with open(f"{self.file_save}/icon_data.csv", 'w+') as icon_output_csv:
            writer = csv.writer(icon_output_csv, delimiter=',')
            writer.writerow([
                'Participant ID',
                dirname(self.file_save).replace(self.data_save_path, '')
            ])
            writer.writerow(['Date/Time', datetime.datetime.now()])
            if self.auc_filename:
                writer.writerow([
                    'Calibration AUC',
                    basename(self.auc_filename).replace('.pkl', '')
                ])

            current_trial = 1 if selections == 0 else selections

            writer.writerow([
                'Percentage of correctly selected icons',
                (correct_trials / current_trial) * 100
            ])
            writer.writerow(['Task type', self.session_description])

    def exit_display(self):
        """Close the UI and cleanup"""
        # Say Goodbye!
        self.rsvp.text = trial_complete_message(self.window, self.parameters)
        self.rsvp.draw_static()
        self.window.flip()

        # Give the system time to process
        core.wait(self.buffer_val)

        if self.daq.is_calibrated:
            write_triggers_from_sequence_icon_to_icon(
                sequence_timing=['offset', self.daq.offset],
                trigger_file=self.trigger_file,
                target=None,
                target_displayed=False,
                offset=True)
        # Close this sessions trigger file and return some data
        self.trigger_file.close()

        # Wait some time before exiting so there is trailing eeg data saved
        core.wait(self.eeg_buffer)

    def init_copy_phrase_task(self, task_list: List[Tuple]):
        """Initialize the CopyPhraseWrapper
        Parameters:
        -----------
            task_list: List of (sequence to match, items matched so far).
        Returns:
        --------
            initialized CopyPhraseWrapper
        """
        return CopyPhraseWrapper(
            self.min_num_seq,
            self.max_seq_per_trial,
            signal_model=self.signal_model,
            fs=self.daq.device_info.fs,
            k=2,
            alp=self.alp,
            task_list=task_list,
            lmodel=self.language_model,
            is_txt_stim=self.is_txt_stim,
            device_name=self.daq.device_info.name,
            device_channels=self.daq.device_info.channels,
            decision_threshold=self.parameters['decision_threshold'],
            stimuli_timing=self.timing[1:])  # time_cross and time_flash

    def stoppage_criteria_ok(self, total_sequences, total_time) -> bool:
        """Returns True if experiment is currently within params, False if
        total sequences or total time exceeds configured values."""
        if total_sequences >= self.max_seq_length:
            self.logger.debug('Max tries exceeded: to allow for more tries'
                              ' adjust the Maximum Sequence Length '
                              '(max_seq_len) parameter.')
            return False

        if total_time >= self.max_seconds:
            self.logger.debug('Max time exceeded. To allow for more time '
                              'adjust the max_minutes parameter.')
            return False
        return True

    def update_session_data(self, data, epoch, save=True):
        """Update the session data and optionally save."""
        data['total_time_spent'] = self.experiment_clock.getTime()
        data['total_number_epochs'] = epoch.epoch_counter
        data['epochs'] = epoch.data
        if save:
            _save_session_related_data(self.session_save_location, data)

    def execute(self):
        self.logger.debug('Starting Icon to Icon Task!')

        icons = [random.choice(self.alp) for _ in range(self.stim_number)]
        self.logger.debug(f'Icon sequence: {icons}')

        selections = []
        copy_phrase_task = self.init_copy_phrase_task(task_list=[(icons, [])])
        epoch = EpochManager(icons, copy_phrase_task, self.timing[0])
        correct_trials = 0

        data = self.init_session_data(save=True)
        run = self.await_start()

        epoch.next()
        while run and self.user_wants_to_continue():

            sequence_timing = self.present_sequence(
                epoch.current_sequence,
                epoch.current_durations,
                show_target=epoch.first_sequence)

            # Write triggers to file
            write_triggers_from_sequence_icon_to_icon(
                sequence_timing, self.trigger_file, epoch.target,
                target_displayed=epoch.first_sequence)

            core.wait(self.buffer_val)

            # Delete calibration
            if epoch.first_stimulus:
                del sequence_timing[0]

            # Delete the target presentation
            if epoch.first_sequence:
                del sequence_timing[0]

            # Reshape the data and triggers as needed for analysis.
            raw_data, triggers, target_info = process_data_for_decision(
                sequence_timing, self.daq, self.window, self.parameters,
                self.rsvp.first_stim_time)

            # Data Record for session.json
            entry = {
                'stimuli': epoch.stimuli,
                'eeg_len': len(raw_data),
                'timing_sti': epoch.timing_sti,
                'triggers': triggers,
                'target_info': target_info,
                'target_letter': epoch.target
            }

            if self.fake:
                new_epoch = True
                correct_trials += 1
                selections.append(epoch.target)
                self.display_feedback(selection=epoch.target, correct=True)
            else:
                # Evaluate the data and make a decision.
                decision_made, new_sti = copy_phrase_task.evaluate_sequence(
                    raw_data, triggers, target_info,
                    self.collection_window_len)
                new_epoch = decision_made

                # Add the evidence to the data record.
                ev_hist = copy_phrase_task.conjugator.evidence_history
                likelihood = copy_phrase_task.conjugator.likelihood
                entry['lm_evidence'] = ev_hist['LM'][0].tolist()
                entry['eeg_evidence'] = ev_hist['ERP'][-1].tolist()
                entry['likelihood'] = likelihood.tolist()

                if decision_made:
                    selection = copy_phrase_task.decision_maker.last_selection
                    selections.append(selection)
                    correct = selection == epoch.target
                    if correct:
                        correct_trials += 1
                    self.display_feedback(selection, correct)

            epoch.add_stim_data(entry)
            self.update_session_data(data, epoch, save=True)
            if new_epoch:
                epoch.next()
            else:
                epoch.next_stimulus(new_sti)

            run = self.stoppage_criteria_ok(epoch.seq_counter,
                                            data['total_time_spent'])
            # end while loop

        self.write_data(correct_trials, len(selections))
        self.exit_display()
        return self.file_save

    @classmethod
    def label(cls):
        return RSVPIconToIconTask.TASK_NAME

    def name(self):
        return RSVPIconToIconTask.TASK_NAME


def _init_icon_to_icon_display_task(parameters, win, daq, static_clock,
                                    experiment_clock, is_word):
    return IconToIconDisplay(
        win,
        static_clock,
        experiment_clock,
        daq.marker_writer,
        info_text=parameters['info_text'],
        info_color=parameters['info_color'],
        info_pos=(
            parameters['text_pos_x'],
            parameters['text_pos_y']),
        info_height=parameters['info_height'],
        info_font=parameters['info_font'],
        task_color=['black'],
        task_font=parameters['task_font'],
        task_height=parameters['task_height'],
        stim_font=parameters['stim_font'],
        stim_pos=(
            parameters['stim_pos_x'],
            parameters['stim_pos_y']),
        stim_height=parameters['stim_height'],
        stim_sequence=['a'] * 10,
        stim_colors=[parameters['stim_color']] * 10,
        stim_timing=[3] * 10,
        is_txt_stim=False,
        trigger_type=parameters['trigger_type'],
        is_word=is_word)


@auto_str
class EpochManager():
    """Manages the state required for tracking epochs.
    Parameters:
    -----------
        icons - sequence of icons to match
        copy_phrase_task - CopyPhraseWrapper
        time_target - time to present a target stimulus.
    """
    # TODO: rather than pass in copy_phrase_task, provide a function to generate
    # stimulus for new epochs.
    def __init__(self, icons, copy_phrase_task, time_target):
        self.copy_phrase_task = copy_phrase_task
        self.target = None
        self.time_target = time_target
        self.icons = icons

        self.current_sequence = []
        self.current_durations = []

        self.epoch_counter = -1
        self.epoch_index = 0
        self.data = {}
        self.seq_counter = 0

    #  Public API
    def next(self):
        """Initialize the next epoch/target"""
        self._increment_epoch()
        self._new_epoch_stimulus()
        self.seq_counter += 1

    def next_stimulus(self, sti):
        """Add another sequence for the current target.
        Parameters:
        -----------
            sti - tuple of (List[alphabet], List[durations], List[colors])
        """
        self.epoch_index += 1
        self.seq_counter += 1
        self._set_stimulus(sti)

    @property
    def first_sequence(self) -> bool:
        """Is the current sequence the first one displayed for this epoch?"""
        return self.epoch_index == 0

    @property
    def first_stimulus(self) -> bool:
        """Is this the first stimulus presented?"""
        return self.epoch_index == 0 and self.first_sequence

    @property
    def stimuli(self) -> List[str]:
        """Stimuli to write to the session.json; excludes the target
        presentation"""
        if self.first_sequence:
            return self.current_sequence[1:]
        else:
            return self.current_sequence

    @property
    def timing_sti(self) -> List[float]:
        """Timing Stimuli to write to the session.json; excludes the target
        presentation"""
        if self.first_sequence:
            return self.current_durations[1:]
        else:
            return self.current_durations

    def add_stim_data(self, data):
        """Add stimulus data for the current epoch_counter/index"""
        self.data[self.epoch_counter][self.epoch_index] = data

    # Internally used methods
    def _increment_epoch(self):
        """Increment the counters and select a new target."""
        self.epoch_counter += 1
        self.epoch_index = 0
        self.target = self.icons[self.epoch_counter]
        self.data[self.epoch_counter] = {}

    def _new_epoch_stimulus(self):
        """Init an epoch, generate initial stimuli seq."""
        _, sti = self.copy_phrase_task.initialize_epoch()
        self._set_stimulus(sti)
        self._new_sequence_mods()

    def _set_stimulus(self, sti: Tuple):
        """Sets the current sequence and current durations from the sti output
        from copy_phrase_task.
        Parameters:
        -----------
            sti - tuple of (List[alphabet], List[durations], List[colors])
        """
        self.current_sequence = sti[0][0]
        self.current_durations = sti[1][0]

    def _new_sequence_mods(self):
        """Make any modifications to the current sequence"""
        # Insert target for initial display
        self.current_sequence.insert(0, self.target)
        self.current_durations.insert(0, self.time_target)



import glob
import itertools
import logging
import random
from os import path, sep

from typing import Iterator

import numpy as np
import sounddevice as sd
import soundfile as sf
from PIL import Image
from psychopy import core

# Prevents pillow from filling the console with debug info
logging.getLogger("PIL").setLevel(logging.WARNING)


def best_selection(selection_elements: list,
                   val: list,
                   len_query: int,
                   always_included=None) -> list:
    """Best Selection.

     given set of elements and a value function over the set, picks the len_query
        number of elements with the best value.

        Args:
            selection_elements(list[str]): the set of elements
            val(list[float]): values for the corresponding elements
            len_query(int): number of elements to be picked from the set
            always_included(list[str]): subset of elements that should always be
                included in the result. Defaults to None.
        Return:
            best_selection(list[str]): elements from selection_elements with the best values """

    always_included = always_included or []
    n = len_query
    # pick the top n items sorted by value in decreasing order
    elem_val = dict(zip(selection_elements, val))
    best = sorted(selection_elements, key=elem_val.get, reverse=True)[0:n]

    replacements = [
        item for item in always_included
        if item not in best and item in selection_elements
    ][0:n]

    if replacements:
        best[-len(replacements):] = replacements
    return best


def best_case_rsvp_seq_gen(alp: list,
                           session_stimuli: list,
                           timing=[1, 0.2],
                           color=['red', 'white'],
                           stim_number=1,
                           stim_length=10,
                           is_txt=True,
                           seq_constants=None) -> tuple:
    """Best Case RSVP Sequence Generation.

    generates RSVPKeyboard sequence by picking n-most likely letters.
        Args:
            alp(list[str]): alphabet (can be arbitrary)
            session_stimuli(ndarray[float]): quantifier metric for query selection
                dim(session_stimuli) = card(alp)!
            timing(list[float]): Task specific timing for generator
            color(list[str]): Task specific color for generator
                First element is the target, second element is the fixation
                Observe that [-1] element represents the trial information
            stim_number(int): number of random stimuli to be created
            stim_length(int): number of trials in a sequence
            seq_constants(list[str]): list of letters that should always be
                included in every sequence. If provided, must be alp items.
        Return:
            schedule_seq(tuple(
                samples[list[list[str]]]: list of sequences
                timing(list[list[float]]): list of timings
                color(list(list[str])): list of colors)): scheduled sequences
            """

    if len(alp) != len(session_stimuli):
        raise Exception('Missing information about alphabet. len(alp):{}, '
                        'len(session_stimuli):{}, should be same!'.format(
                            len(alp), len(session_stimuli)))

    if seq_constants and not set(seq_constants).issubset(alp):
        raise Exception('Sequence constants must be alphabet items.')

    # create a list of alphabet letters
    alphabet = [i for i in alp]

    # query for the best selection
    query = best_selection(alphabet, session_stimuli, stim_length, seq_constants)

    # shuffle the returned values
    random.shuffle(query)

    # Init some lists to construct our stimuli with
    samples, times, colors = [], [], []
    for idx_num in range(stim_number):

        # append a fixation cross. if not text, append path to image fixation
        if is_txt:
            sample = ['+']
        else:
            sample = ['bcipy/static/images/bci_main_images/PLUS.png']

        # construct the sample from the query
        sample += [i for i in query]
        samples.append(sample)

        # append timing
        times.append([timing[i] for i in range(len(timing) - 1)] +
                     [timing[-1]] * stim_length)

        # append colors
        colors.append([color[i] for i in range(len(color) - 1)] +
                      [color[-1]] * stim_length)
    return (samples, times, colors)


def random_rsvp_calibration_seq_gen(alp, timing=[0.5, 1, 0.2],
                                    color=['green', 'red', 'white'],
                                    stim_number=10,
                                    stim_length=10, is_txt=True):
    """Random RSVP Calibration Sequence Generator.

    Generates random RSVPKeyboard sequences.
        Args:
            alp(list[str]): alphabet (can be arbitrary)
            timing(list[float]): Task specific timing for generator
            color(list[str]): Task specific color for generator
                First element is the target, second element is the fixation
                Observe that [-1] element represents the trial information
            stim_number(int): number of random stimuli to be created
            stim_length(int): number of trials in a sequence
        Return:
            schedule_seq(tuple(
                samples[list[list[str]]]: list of sequences
                timing(list[list[float]]): list of timings
                color(list(list[str])): list of colors)): scheduled sequences
    """

    len_alp = len(alp)

    samples, times, colors = [], [], []
    for idx_num in range(stim_number):
        idx = np.random.permutation(np.array(list(range(len_alp))))
        rand_smp = (idx[0:stim_length])
        if not is_txt:
            sample = [
                alp[rand_smp[0]],
                'bcipy/static/images/bci_main_images/PLUS.png']
        else:
            sample = [alp[rand_smp[0]], '+']
        rand_smp = np.random.permutation(rand_smp)
        sample += [alp[i] for i in rand_smp]
        samples.append(sample)
        times.append([timing[i] for i in range(len(timing) - 1)] +
                     [timing[-1]] * stim_length)
        colors.append([color[i] for i in range(len(color) - 1)] +
                      [color[-1]] * stim_length)

    schedule_seq = (samples, times, colors)

    return schedule_seq


def target_rsvp_sequence_generator(alp, target_letter, parameters, timing=[0.5, 1, 0.2],
                                   color=['green', 'white', 'white'],
                                   stim_length=10, is_txt=True):
    """Target RSVP Sequence Generator.

    Generate target RSVPKeyboard sequences.

        Args:
            alp(list[str]): alphabet (can be arbitrary)
            target_letter([str]): letter to be copied
            timing(list[float]): Task specific timing for generator
            color(list[str]): Task specific color for generator
                First element is the target, second element is the fixation
                Observe that [-1] element represents the trial information
            stim_length(int): number of trials in a sequence
        Return:
            schedule_seq(tuple(
                samples[list[list[str]]]: list of sequences
                timing(list[list[float]]): list of timings
                color(list(list[str])): list of colors)): scheduled sequences
    """

    len_alp = len(alp)

    # intialize our arrays
    samples, times, colors = [], [], []
    rand_smp = random.sample(range(len_alp), stim_length)
    if is_txt:
        sample = ['+']
    else:
        sample = ['bcipy/static/images/bci_main_images/PLUS.png']
        target_letter = parameters['path_to_presentation_images'] + \
            target_letter + '.png'
    sample += [alp[i] for i in rand_smp]

    # if the target isn't in the array, replace it with some random index that
    #  is not fixation
    if target_letter not in sample:
        random_index = np.random.randint(0, stim_length - 1)
        sample[random_index + 1] = target_letter

    # add target letter to start
    sample = [target_letter] + sample

    # to-do shuffle the target letter

    samples.append(sample)
    times.append([timing[i] for i in range(len(timing) - 1)] +
                 [timing[-1]] * stim_length)
    colors.append([color[i] for i in range(len(color) - 1)] +
                  [color[-1]] * stim_length)

    schedule_seq = (samples, times, colors)

    return schedule_seq


def get_task_info(experiment_length, task_color):
    """Get Task Info.

    Generates fixed RSVPKeyboard task text and color information for
            display.
    Args:
        experiment_length(int): Number of sequences for the experiment
        task_color(str): Task information display color

    Return get_task_info((tuple): task_text: array of task text to display
                   task_color: array of colors for the task text
                   )
    """

    # Do list comprehensions to get the arrays for the task we need.

    task_text = ['%s/%s' % (stim + 1, experiment_length)
                 for stim in range(experiment_length)]
    task_color = [[str(task_color)] for stim in range(experiment_length)]

    return (task_text, task_color)


def rsvp_copy_phrase_seq_generator(alp, target_letter, timing=[0.5, 1, 0.2],
                                   color=['green', 'white', 'white'],
                                   stim_length=10):
    """Generate copy phrase RSVPKeyboard sequences.

        Args:
            alp(list[str]): alphabet (can be arbitrary)
            target_letter([str]): letter to be copied
            timing(list[float]): Task specific timing for generator
            color(list[str]): Task specific color for generator
                First element is the target, second element is the fixation
                Observe that [-1] element represents the trial information
            stim_length(int): number of trials in a sequence
        Return:
            schedule_seq(tuple(
                samples[list[list[str]]]: list of sequences
                timing(list[list[float]]): list of timings
                color(list(list[str])): list of colors)): scheduled sequences
    """

    len_alp = len(alp)

    # initialize our arrays
    samples, times, colors = [], [], []
    rand_smp = np.random.randint(0, len_alp, stim_length)
    sample = ['+']
    sample += [alp[i] for i in rand_smp]

    # if the target isn't in the array, replace it with some random index that
    #  is not fixation
    if target_letter not in sample:
        random_index = np.random.randint(0, stim_length - 1)
        sample[random_index + 1] = target_letter

    # to-do shuffle the target letter

    samples.append(sample)
    times.append([timing[i] for i in range(len(timing) - 1)] +
                 [timing[-1]] * stim_length)
    colors.append([color[i] for i in range(len(color) - 1)] +
                  [color[-1]] * stim_length)

    schedule_seq = (samples, times, colors)

    return schedule_seq


def generate_icon_match_images(experiment_length, image_path, number_of_sequences, timing, is_word):
    """Generate Image Icon Matches.

    Generates an array of images to use for the icon matching task.
    Args:
        experiment_length(int): Number of images per sequence
        image_path(str): Path to image files
        number_of_sequences(int): Number of sequences to generate
        timing(list): List of timings; [parameters['time_target'],
                       parameters['time_cross'],
                       parameters['time_flash']]
        is_word(bool): Whether or not this is an icon to word matching task
    Return generate_icon_match_images(arrays of tuples of paths to images to
    display, and timings)
    """
    # Get all png images in image path
    image_array = [
        img for img in glob.glob(image_path + "*.png")
        if not img.endswith("PLUS.png")
    ]

    if experiment_length > len(image_array) - 1:
        raise Exception(
            'Number of images to be displayed on screen is longer than number of images available')

    # Generate indexes of target images
    target_image_numbers = np.random.randint(
        0, len(image_array), number_of_sequences)

    # Array of images to return
    return_array = []

    # Array of timings to return
    return_timing = []
    for specific_time in range(len(timing) - 1):
        return_timing.append(timing[specific_time])
    for item_without_timing in range(len(return_timing), experiment_length):
        return_timing.append(timing[-1])

    for sequence in range(number_of_sequences):
        return_array.append([])
        # Generate random permutation of image indexes
        random_number_array = np.random.permutation(len(image_array))
        if is_word:
            # Add name of target image to array
            image_path = path.basename(
                image_array[target_image_numbers[sequence]])
            return_array[sequence].append(image_path.replace('.png', ''))
        else:
            # Add target image to image array
            return_array[sequence].append(
                image_array[target_image_numbers[sequence]])
        # Add PLUS.png to image array
        return_array[sequence].append(
            'bcipy/static/images/bci_main_images/PLUS.png')

        # Add target image to sequence, if it is not already there
        if not target_image_numbers[sequence] in random_number_array[2:experiment_length]:
            random_number_array[np.random.randint(
                2, experiment_length)] = target_image_numbers[sequence]

        # Fill the rest of the image array with random images
        for item in range(2, experiment_length):
            return_array[sequence].append(
                image_array[random_number_array[item]])

    return (return_array, return_timing)


def resize_image(image_path: str, screen_size: tuple, sti_height: int):
    """Resize Image.

    Returns the width and height that a given image should be displayed at
    given the screen size, size of the original image, and stimuli height
    parameter"""
    # Retrieve image width and height
    with Image.open(image_path) as pillow_image:
        image_width, image_height = pillow_image.size

    # Resize image so that its largest dimension is the stimuli size defined in the parameters file
    if image_width >= image_height:
        proportions = (1, (image_height / image_width))
    else:
        proportions = ((image_width / image_height), 1)

    # Adjust image size to scale with monitor size
    screen_width, screen_height = screen_size
    if screen_width >= screen_height:
        sti_size = ((screen_height / screen_width) * sti_height *
                    proportions[0],  sti_height * proportions[1])
    else:
        sti_size = (
            sti_height * proportions[0], (screen_width / screen_height) * sti_height * proportions[1])

    return sti_size


def play_sound(sound_file_path: str,
               dtype: str='float32',
               track_timing: bool=False,
               sound_callback=None,
               sound_load_buffer_time: float=0.5,
               sound_post_buffer_time: float=1,
               experiment_clock=None,
               trigger_name: str=None,
               timing: list=[]) -> list:
    """Play Sound.

    Using soundevice and soundfile, play a sound giving options to buffer times between
        loading sound into memory and after playing. If desired, marker writers or list based
        timing with psychopy clocks may be passed and sound timing returned.


    PARAMETERS
    ----------
    :param: sound_file_path
    :param: dtype: type of sound ex. float32.
    :param: track_timing: whether or not to track timing of sound playin
    :param: sound_callback: trigger based callback (see MarkerWriter and NullMarkerWriter)
    :param: sound_load_buffer_time: time to wait after loading file before playing
    :param: sound_post_buffer_time: time to wait after playing sound before returning
    :param: experiment_clock: psychopy clock to get time of sound stimuli
    :param: trigger_name: name of the sound trigger
    :param: timing: list of triggers in the form of trigger name, trigger timing
    :resp: timing
    """

    try:
        # load in the sound file and wait some time before playing
        data, fs = sf.read(
            sound_file_path, dtype=dtype)
        core.wait(sound_load_buffer_time)

    except Exception:
        raise Exception(
            'StimGenPlaySoundError: sound file could not be found or initialized.')

    #  if timing is wanted, get trigger timing for this sound stimuli
    if track_timing:
        timing.append(trigger_name)
        timing.append(experiment_clock.getTime())

        # if there is a timing callback for sound, evoke it with the timing list
        if sound_callback is not None:
            sound_callback(timing)

    # play our loaded sound and wait for some time before it's finished
    # NOTE: there is a measurable delay for calling sd.play. (~ 0.1 seconds;
    # which I believe happens prior to the sound playing).
    sd.play(data, fs)
    if sound_post_buffer_time:
        # sd.play returns immediately (according to the docs); calculate offset
        # so the sound_post_buffer_time accounts for the duration of the sound.
        duration = len(data) / fs
        core.wait(sound_post_buffer_time + duration)
    return timing


def soundfiles(directory: str) -> Iterator[str]:
    """Creates a generator that cycles through sound files (.wav) in a 
    directory and returns the path to next sound file on each iteration.

    Parameters:
    -----------
        directory - path to the directory which contains .wav files
    Returns:
    --------
        iterator that infinitely cycles through the filenames.
    """
    if not path.isdir(directory):
        raise Exception(('Invalid directory for sound files. Please check '
                         'your configuration.'))
    if not directory.endswith(sep):
        directory += sep
    return itertools.cycle(glob.glob(directory + '*.wav'))



import numpy as np
from scipy.signal import filtfilt, iirnotch


def notch_filter(data: np.ndarray, fs: int, frequency_to_remove: int, quality_factor: int=30) -> np.ndarray:
    """Notch Filter.

    A notch filter is a bandstop filter with a narrow bandwidth. It removes
        the frequency of interest without much impact to others.
    """
    normalized_frequency = frequency_to_remove / (fs / 2)
    b, a = iirnotch(normalized_frequency, quality_factor)
    filtered_data = filtfilt(b, a, data)
    return filtered_data



import numpy as np
import scipy.optimize
from sklearn import metrics

import logging

log = logging.getLogger(__name__)


def cost_cross_validation_auc(model, opt_el, x, y, param, k_folds=10,
                              split='uniform'):
    """ Minimize cost of the overall -AUC.
        Cost function: given a particular architecture (model). Fits the
        parameters to the folds with leave one fold out procedure. Calculates
        scores for the validation fold. Concatenates all calculated scores
        together and returns a -AUC vale.
        Args:
            model(pipeline): model to be iterated on
            opt_el(int): number of the element in pipeline to be optimized
            x(ndarray[float]): N x k data array
            y(ndarray[int]): N x k observation (class) array
                N is number of samples k is dimensionality of features
            param(list[float]): ordered hyper parameter list for the
                regularization
            k_folds(int): number of folds
            split(string): split type,
                'uniform': Takes the data as is
        Return:
            -auc(float): negative AUC value for current setup
            sc_h(ndarray[float]): scores computed for each validation fold
            y_valid_h(ndarray[int]): labels of the scores for each validation fold
                y_valid_h[i] is basically the label for sc_h[i] """

    num_samples = x.shape[1]
    fold_len = np.floor(float(num_samples) / k_folds)

    model.pipeline[1].lam = param[0]
    model.pipeline[1].gam = param[1]

    fold_x, fold_y = [], []
    sc_h, y_valid_h = [], []
    if split == 'uniform':
        for idx_fold in range(k_folds + 1):
            fold_x.append(x[:, int(idx_fold * fold_len):int(
                (idx_fold + 1) * fold_len), :])
            fold_y.append(y[int(idx_fold * fold_len):int((idx_fold + 1) *
                                                         fold_len)])
            if len(np.unique(fold_y[idx_fold])) == 1:
                raise Exception('Cannot use {}-folding in cross_validation '
                                'or # of folds is inconsistent'.format(split))

        for idx_fold in range(k_folds):
            list_valid = idx_fold
            list_train = list(set(range(k_folds)) - set([idx_fold]))

            x_train = np.concatenate([fold_x[i] for i in list_train], axis=1)
            y_train = np.concatenate([fold_y[i] for i in list_train], axis=0)
            x_valid = fold_x[list_valid]
            y_valid = fold_y[list_valid]

            model.fit(x_train, y_train)
            sc = model.transform(x_valid)

            sc_h.append(sc)
            y_valid_h.append(y_valid)

    y_valid_h = np.concatenate(np.array(y_valid_h))
    sc_h = np.concatenate(np.array(sc_h))
    fpr, tpr, _ = metrics.roc_curve(y_valid_h, sc_h, pos_label=1)
    auc = metrics.auc(fpr, tpr)

    return -auc, sc_h, y_valid_h


def grid_search(model, opt_el, x, y, grid=[5, 5], op_type='cost_auc',
                arg_op_type=[10, 'uniform']):
    """ Description: This function performs an exhaustive grid search
                     to estimate the hyper parameters lambda and gamma
                     that minimize the cost of AUC.
        Args:
            model(pipeline): model to be iterated on
            opt_el(int): number of the element in pipeline to be optimized
            x(ndarray[float]): N x k data array
            y(ndarray[int]): N x k observation (class) array
                 N is number of samples k is dimensionality of features
            grid(list(int)): a list of 2 numbers for grid
            op_type(string): type of the optimization
        Returns:
            arg_opt(list[float]): optimized hyper parameters
            """
    if op_type == 'cost_auc':
        k_folds, split = arg_op_type
        # This specifies the different candidate values we want to try.
        # The grid search will try all combination of these parameter values
        # and select the set of parameters that provides the most accurate
        # model.
        param_cand = {  # dictionary of parameter candidates
            'lam': np.linspace(.01, .99, grid[0], endpoint=False),
            'gam': np.linspace(.01, .99, grid[1], endpoint=False),
        }
        best_auc = 0  # auc can't be smaller than 0
        arg_opt = {'lam': 0, 'gam': 0}

        # For every coordinate on the grid, try every combination of
        # hyper parameters:
        tmp_counter = 0
        for i in range(len(param_cand['lam'])):
            for j in range(len(param_cand['gam'])):
                auc = -cost_cross_validation_auc(model, opt_el, x, y,
                                                 [param_cand['lam'][i],
                                                  param_cand['gam'][j]],
                                                 k_folds=k_folds, split=split)[0]
                if auc > best_auc:
                    best_auc = auc
                    arg_opt['lam'], arg_opt['gam'] = param_cand['lam'][i], \
                                                     param_cand['gam'][j]

                tmp_counter += 1
    else:
        raise Exception('Error: Operation type other than AUC cost.')

    # This returns the parameter estimates with the highest scores:
    return [arg_opt['lam'], arg_opt['gam']]


def nonlinear_opt(model, opt_el, x, y, init=None, op_type='cost_auc',
                  arg_op_type=[10, 'uniform']):
    """ Optimizes lambda, gamma values for given  penalty function
        Args:
            model(pipeline): model to be iterated on
            opt_el(int): number of the element in pipeline to be optimized
            x(ndarray[float]): N x k data array
            y(ndarray[int]): N x k observation (class) array
                 N is number of samples k is dimensionality of features
            init(list[float]): initial values for gamma and lambda
            op_type(string): type of the optimization
        Return:
            arg_opt(list[float]): optimized hyper parameters
            """

    # Get initial values
    if not init:
        init = [model.pipeline[opt_el].lam, model.pipeline[opt_el].gam]
    if op_type:
        # TODO: maybe we should not have such an option and set it by ourselves
        if op_type == 'cost_auc':
            k_folds, split = arg_op_type
            cost_fun_param = lambda b: cost_cross_validation_auc(
                model, opt_el, x, y, [b[0], b[1]], k_folds=k_folds,
                split=split)[0]

        # Intervals for lambda and gamma parameters
        # Observe that 0 < lam < 1, 0 < gam < 1
        cst_1 = lambda v: v[0] - np.power(0.1, 15)
        cst_2 = lambda v: v[1] - np.power(0.1, 15)
        cst_3 = lambda v: 1 - v[0]
        cst_4 = lambda v: 1 - v[1]

        arg_opt = scipy.optimize.fmin_cobyla(cost_fun_param, x0=init,
                                             disp=False,
                                             cons=[cst_1, cst_2, cst_3,
                                                   cst_4])
    return arg_opt


def cross_validation(x, y, model, opt_el=1, k_folds=10, split='uniform'):
    """ Cross validation function for hyper parameter optimization
        Args:
            x(ndarray[float]): C x N x k data array
            y(ndarray[int]): N x 1 observation (class) array
                N is number of samples k is dimensionality of features
                C is number of channels
            model(pipeline): model to be optimized
            opt_el(int): element in the model to update hyper-params in [0,M]
            k_folds(int): number of folds
            split(string): split type,
                'uniform': Takes the data as is
        Return:
            return(list[list[float],list[float],list[float]]): lists of lambda,
                gamma and AUC values for each fold respectively.
            """

    log.debug('Starting Cross Validation !')
    arg_opt = nonlinear_opt(model, opt_el, x, y, op_type='cost_auc',
                            arg_op_type=[k_folds, split])
    return arg_opt



from psychopy import visual
from bcipy.display.rsvp.display import RSVPDisplay
from bcipy.helpers.stimuli import resize_image
from bcipy.helpers.task import SPACE_CHAR


class CalibrationDisplay(RSVPDisplay):
    """Calibration Display."""
    def __init__(self,
                 window,
                 static_clock,
                 experiment_clock,
                 marker_writer,
                 info_text='Press Space Bar to Pause',
                 info_color='White',
                 info_pos=(0, -.9),
                 info_height=0.2,
                 info_font='Times',
                 task_color=['white'],
                 task_font='Times',
                 task_text='1/100',
                 task_height=0.1,
                 stim_font='Times',
                 stim_pos=(-.8, .9),
                 stim_height=0.2,
                 stim_sequence=['a'] * 10,
                 stim_colors=['white'] * 10,
                 stim_timing=[1] * 10,
                 is_txt_stim=True,
                 trigger_type='image',
                 space_char=SPACE_CHAR):

        info_color = [info_color]
        info_font = [info_font]
        info_text = [info_text]
        info_pos = [info_pos]
        info_height = [info_height]

        tmp = visual.TextStim(win=window, font=task_font, text=task_text)
        x_task_pos = tmp.boundingBox[0] / window.size[0] - 1
        task_pos = (x_task_pos, 1 - task_height)

        super(CalibrationDisplay, self).__init__(
            window,
            static_clock,
            experiment_clock,
            marker_writer,
            task_color=task_color,
            task_font=task_font,
            task_pos=task_pos,
            task_height=task_height,
            task_text=task_text,
            info_color=info_color,
            info_text=info_text,
            info_font=info_font,
            info_pos=info_pos,
            stim_font=stim_font,
            stim_pos=stim_pos,
            stim_height=stim_height,
            stim_sequence=stim_sequence,
            stim_colors=stim_colors,
            stim_timing=stim_timing,
            is_txt_stim=is_txt_stim,
            trigger_type=trigger_type,
            space_char=space_char)



"""Code for mocking an EEG data stream. Code in this module produces data
at a specified frequency."""
import logging
from builtins import next
from queue import Queue
import random
import threading
import time

from bcipy.acquisition.datastream.generator import random_data

log = logging.getLogger(__name__)


class Producer(threading.Thread):
    """Produces generated data at a specified frequency.

    Parameters
    ----------
        queue : Queue
            Generated data will be written to the queue.
        freq : float, optional
            Data will be generated at the given frequency.
        generator : object, optional
            python generator for creating data.
        maxiters : int, optional
            if provided, stops generating data after the given number of iters.
    """

    def __init__(self,
                 queue,
                 freq=1 / 100,
                 generator=random_data(),
                 maxiters=None):

        super(Producer, self).__init__()
        self.daemon = True
        self._running = True

        self.freq = freq
        self.generator = generator
        self.maxiters = maxiters
        self.queue = queue

    # @override to make this class a context manager
    def __enter__(self):
        self.start()
        return self

    # @override to make this class a context manager
    def __exit__(self, _exc_type, _exc_value, _traceback):
        self.stop()

    def _genitem(self):
        """Generates the data item to be added to the queue."""

        try:
            data = next(self.generator)
        except StopIteration:
            log.debug("End of input reached")
            raise Exception("End of input reached")
        return data

    def _additem(self):
        """Adds the data item to the queue."""

        self.queue.put(self._genitem())

    def stop(self):
        """Stop the thread; stopped threads cannot be restarted."""

        self._running = False

    def run(self):
        """Provides a control loop, adding a data item to the queue at the
        configured frequency.

        @overrides the Thread run method
        """

        def tick():
            """Corrects the time interval if the time of the work to add the
            item causes drift."""
            current_time = time.time()
            count = 0
            while True:
                count += 1
                yield max(current_time + count * self.freq - time.time(), 0)

        sleep_len = tick()
        times = 0
        while self._running and (self.maxiters is None
                                 or times < self.maxiters):
            times += 1
            time.sleep(next(sleep_len))
            self._additem()


class _ConsumerThread(threading.Thread):
    """Consumer used to test the Producer by consuming generated items."""

    def __init__(self, queue, name=None):
        super(_ConsumerThread, self).__init__()
        self.daemon = True
        self.name = name
        self._q = queue

    def run(self):
        while True:
            if not self._q.empty():
                item = self._q.get()
                log.debug('Getting %s: %s items in queue',
                              str(item), str(self._q.qsize()))
                time.sleep(random.random())


def main():
    """Main method"""
    data_queue = Queue()
    producer = Producer(data_queue)
    consumer = _ConsumerThread(data_queue)

    producer.start()
    consumer.start()
    time.sleep(5)


if __name__ == '__main__':
    main()



"""Defines the Protocol data structure"""
from collections import namedtuple

Protocol = namedtuple(
    'Protocol', ['encoder', 'init_messages', 'fs', 'channels'])
Protocol.__doc__ = ('Protocols are primarily used for generating data for '
                    'testing purposes')



# -*- coding: utf-8 -*-

from bcipy.helpers.stimuli import best_case_rsvp_seq_gen
from bcipy.helpers.task import SPACE_CHAR
from typing import Dict, List
import logging
import numpy as np
import string

log = logging.getLogger(__name__)

class EvidenceFusion(object):
    """ Fuses likelihood evidences provided by the inference
        Attr:
            evidence_history(dict{list[ndarray]}): Dictionary of difference
                evidence types in list. Lists are ordered using the arrival
                time.
            likelihood(ndarray[]): current probability distribution over the
                set. Gets updated once new evidence arrives. """

    def __init__(self, list_name_evidence, len_dist):
        self.evidence_history = {name: [] for name in list_name_evidence}
        self.likelihood = np.ones(len_dist) / len_dist

    def update_and_fuse(self, dict_evidence):
        """ Updates the probability distribution
            Args:
                dict_evidence(dict{name: ndarray[float]}): dictionary of
                    evidences (EEG and other likelihoods)
        """

        for key in dict_evidence.keys():
            tmp = dict_evidence[key][:][:]
            self.evidence_history[key].append(tmp)

        # Current rule is to multiply
        for value in dict_evidence.values():
            self.likelihood *= value[:]

        # I
        if np.isinf(np.sum(self.likelihood)):
            tmp = np.zeros(len(self.likelihood))
            tmp[np.where(self.likelihood == np.inf)[0][0]] = 1
            self.likelihood = tmp

        if not np.isnan(np.sum(self.likelihood)):
            self.likelihood = self.likelihood / np.sum(self.likelihood)

        likelihood = self.likelihood[:]

        return likelihood

    def reset_history(self):
        """ Clears evidence history """
        for value in self.evidence_history.values():
            del value[:]
        self.likelihood = np.ones(len(self.likelihood)) / len(self.likelihood)

    def save_history(self):
        """ Saves the current likelihood history """
        return 0


# Criteria
class DecisionCriteria():
    """Abstract class for Criteria which can be applied to evaluate a sequence
    """

    def apply(self, epoch, commit_params):
        """
        Apply the given criteria.
        Parameters:
        -----------
            epoch - Epoch data
                - target(str): target of the epoch
                - time_spent(ndarray[float]): |num_trials|x1
                      time spent on the sequence
                - list_sti(list[list[str]]): presented symbols in each
                      sequence
                - list_distribution(list[ndarray[float]]): list of |alp|x1
                        arrays with prob. dist. over alp
            commit_params - params relevant to stoppage criteria
                min_num_seq: int - minimum number of sequences required
                max_num_seq: int - max number of sequences allowed
                threshold: float - minimum likelihood required
        """
        raise NotImplementedError()


class MinIterationsCriteria(DecisionCriteria):
    """Returns true if the minimum number of iterations have not yet been reached."""

    def apply(self, epoch, commit_params):
        current_seq = len(epoch['list_distribution'])
        return current_seq < commit_params['min_num_seq']


class DecreasedProbabilityCriteria(DecisionCriteria):
    """Returns true if the letter with the max probability decreased from the
        last sequence."""

    def apply(self, epoch, commit_params):
        if len(epoch['list_distribution']) < 2:
            return False
        prev_dist = epoch['list_distribution'][-2]
        cur_dist = epoch['list_distribution'][-1]
        return np.argmax(cur_dist) == np.argmax(
            prev_dist) and np.max(cur_dist) < np.max(prev_dist)


class MaxIterationsCriteria(DecisionCriteria):
    """Returns true if the max iterations have been reached."""

    def apply(self, epoch, commit_params):
        current_seq = len(epoch['list_distribution'])
        if current_seq >= commit_params['max_num_seq']:
            log.debug(
                "Committing to decision: max iterations have been reached.")
            return True
        return False


class CommitThresholdCriteria(DecisionCriteria):
    """Returns true if the commit threshold has been met."""

    def apply(self, epoch, commit_params):
        current_distribution = epoch['list_distribution'][-1]
        if np.max(current_distribution) > commit_params['threshold']:
            log.debug("Committing to decision: Likelihood exceeded threshold.")
            return True
        return False


class CriteriaEvaluator():
    """Evaluates whether an epoch should commit to a decision based on the
    provided criteria.

    Parameters:
    -----------
        continue_criteria: list of criteria; if any of these evaluate to true the
            decision maker continues.
        commit_criteria: list of criteria; if any of these return true and
            continue_criteria are all false, decision maker commits to a decision.
    """

    def __init__(self, continue_criteria: List[DecisionCriteria],
                 commit_criteria: List[DecisionCriteria]):
        self.continue_criteria = continue_criteria or []
        self.commit_criteria = commit_criteria or []

    @classmethod
    def default(cls):
        return cls(continue_criteria=[MinIterationsCriteria()],
                   commit_criteria=[
                       MaxIterationsCriteria(),
                       CommitThresholdCriteria()
                   ])

    def should_commit(self, epoch: Dict, params: Dict):
        """Evaluates the given epoch; returns true if stoppage criteria has
        been met, otherwise false.

        Parameters:
        -----------
            epoch - Epoch data
                - target(str): target of the epoch
                - time_spent(ndarray[float]): |num_trials|x1
                      time spent on the sequence
                - list_sti(list[list[str]]): presented symbols in each
                      sequence
                - list_distribution(list[ndarray[float]]): list of |alp|x1
                        arrays with prob. dist. over alp
            params - params relevant to stoppage criteria
                min_num_seq: int - minimum number of sequences required
                max_num_seq: int - max number of sequences allowed
                threshold: float - minimum likelihood required
        """
        if any(
                criteria.apply(epoch, params)
                for criteria in self.continue_criteria):
            return False
        return any(
            criteria.apply(epoch, params) for criteria in self.commit_criteria)


class DecisionMaker:
    """ Scheduler of the entire framework
        Attr:
            decision_threshold: Minimum combined likelihood required for a
                decision
            state(str): state of the framework, which increases in size
                by 1 after each sequence. Elements are alphabet, ".,_,<"
                where ".": null_sequence(no decision made)
                      "_": space bar
                      "<": back space
            displayed_state(str): visualization of the state to the user
                only includes the alphabet and "_"
            alphabet(list[str]): list of symbols used by the framework. Can
                be switched with location of images or one hot encoded images.
            time(float): system time
            evidence(list[str]): list of evidences used in the framework
            list_priority_evidence(list[]): priority list for the evidences
            sequence_counter(dict[str(val=float)]): number of sequences
                passed for each particular evidence
            list_epoch(list[epoch]): List of stimuli in each sequence
                epoch(dict{items}):
                    - target(str): target of the epoch
                    - time_spent(ndarray[float]): |num_trials|x1
                      time spent on the sequence
                    - list_sti(list[list[str]]): presented symbols in each
                      sequence
                    - list_distribution(list[ndarray[float]]): list of |alp|x1
                        arrays with prob. dist. over alp
            seq_constants(list[str]): list of letters which should appear in
                every sequence.
            criteria_evaluator: CriteriaEvaluator - optional parameter to
                provide alternative rules for committing to a decision.

        Functions:
            decide():
                Checks the criteria for making and epoch, using all
                evidences and decides to do an epoch or to collect more
                evidence
            do_epoch():
                Once committed an epoch perform updates to condition the
                distribution on the previous letter.
            schedule_sequence():
                schedule the next sequence using the current information
            decide_state_update():
                If committed to an epoch update the state using a decision
                metric.
                (e.g. pick the letter with highest likelihood)
            prepare_stimuli():
                prepares the query set for the next sequence
                (e.g pick n-highest likely letters and randomly shuffle)
        """

    def __init__(self,
                 min_num_seq,
                 max_num_seq,
                 decision_threshold=0.8,
                 state='',
                 alphabet=list(string.ascii_uppercase) + ['<'] + [SPACE_CHAR],
                 is_txt_stim=True,
                 stimuli_timing=[1, .2],
                 seq_constants=None,
                 criteria_evaluator=CriteriaEvaluator.default()):
        self.state = state
        self.displayed_state = self.form_display_state(state)
        self.stimuli_timing = stimuli_timing

        # TODO: read from parameters file
        self.alphabet = alphabet
        self.is_txt_stim = is_txt_stim

        self.list_epoch = [{'target': None, 'time_spent': 0,
                            'list_sti': [], 'list_distribution': [], 'decision': None}]
        self.time = 0
        self.sequence_counter = 0

        # Stopping Criteria
        self.min_num_seq = min_num_seq
        self.max_num_seq = max_num_seq

        self.posterior_commit_threshold = decision_threshold

        self.last_selection = ''

        # Items shown in every sequence
        self.seq_constants = seq_constants

        self.criteria_evaluator = criteria_evaluator

    def reset(self, state=''):
        """ Resets the decision maker with the initial state
            Args:
                state(str): current state of the system """
        self.state = state
        self.displayed_state = self.form_display_state(self.state)

        self.list_epoch = [{'target': None, 'time_spent': 0,
                            'list_sti': [], 'list_distribution': []}]
        self.time = 0
        self.sequence_counter = 0

    def form_display_state(self, state):
        """ Forms the state information or the user that fits to the
            display. Basically takes '.' and '<' into consideration and rewrites
            the state
            Args:
                state(str): state string
            Return:
                displayed_state(str): state without '<,.' and removes
                    backspaced letters """
        tmp = ''
        for i in state:
            if i == '<':
                tmp = tmp[0:-1]
            elif i != '.':
                tmp += i

        return tmp

    def update(self, state=''):
        self.state = state
        self.displayed_state = self.form_display_state(state)

    def decide(self, p):
        """ Once evidence is collected, decision_maker makes a decision to
            stop or not by leveraging the information of the stopping
            criteria. Can decide to do an epoch or schedule another sequence.
            Args:
                p(ndarray[float]): |A| x 1 distribution array
                    |A|: cardinality of the alphabet
            Return:
                commitment(bin): True if a letter is a commitment is made
                                 False if requires more evidence
                args(dict[]): Extra arguments depending on the decision
                """

        self.list_epoch[-1]['list_distribution'].append(p[:])

        params = dict(
            min_num_seq=self.min_num_seq,
            max_num_seq=self.max_num_seq,
            threshold=self.posterior_commit_threshold)

        # Check stopping criteria
        if self.criteria_evaluator.should_commit(self.list_epoch[-1], params):
            self.do_epoch()
            return True, []
        else:
            stimuli = self.schedule_sequence()
            return False, {'stimuli': stimuli}


    def do_epoch(self):
        """ Epoch refers to a commitment to a decision.
            If made, state is updated, displayed state is updated
            a new epoch is appended. """
        self.sequence_counter = 0
        decision = self.decide_state_update()
        self.last_selection = decision
        self.state += decision
        self.displayed_state = self.form_display_state(self.state)

        # Initialize next epoch
        self.list_epoch.append({'target': None, 'time_spent': 0,
                                'list_sti': [], 'list_distribution': []})

    def schedule_sequence(self):
        """ Schedules next sequence """
        self.state += '.'
        stimuli = self.prepare_stimuli()
        self.list_epoch[-1]['list_sti'].append(stimuli[0])
        self.sequence_counter += 1

        return stimuli

    def decide_state_update(self):
        """ Checks stopping criteria to commit to an epoch """
        idx = np.where(
            self.list_epoch[-1]['list_distribution'][-1] ==
            np.max(self.list_epoch[-1]['list_distribution'][-1]))[0][0]
        decision = self.alphabet[idx]
        self.list_epoch[-1]['decision'] = decision
        return decision

    def prepare_stimuli(self):
        """ Given the alphabet, under a rule, prepares a stimuli for
            the next sequence
            Return:
                stimuli(tuple[list[char],list[float],list[str]]): tuple of
                    stimuli information. [0]: letter, [1]: timing, [2]: color
                """

        stimuli = best_case_rsvp_seq_gen(
            self.alphabet,
            self.list_epoch[-1]['list_distribution'][-1],
            stim_number=1,
            is_txt=self.is_txt_stim,
            timing=self.stimuli_timing,
            seq_constants=self.seq_constants)
        return stimuli



"""Tests for DSI headset device driver."""
import unittest
import pytest
from bcipy.acquisition.datastream import server
from bcipy.acquisition.protocols.dsi.dsi_device import dsi
from bcipy.acquisition.protocols.dsi.dsi_device import DsiDevice
from bcipy.acquisition.protocols.dsi.dsi_protocol import DsiProtocol


HOST = '127.0.0.1'
DEFAULT_PORT = 9000


class TestDsiDevice(unittest.TestCase):
    """Tests for DsiDevice"""

    def __init__(self, *args, **kwargs):
        super(TestDsiDevice, self).__init__(*args, **kwargs)
        self.host = HOST

    def connection_params(self):
        """Return connection params dict"""
        return {'host': self.host, 'port': self.port}

    @classmethod
    def setUpClass(cls):
        cls.server, cls.port = server.start_socket_server(
            DsiProtocol(), HOST, DEFAULT_PORT)

    @classmethod
    def tearDownClass(cls):
        cls.server.stop()

    def setUp(self):
        """Run before each test."""

        self.port = type(self).port

    def test_mismatched_channels(self):
        """An exception should be thrown if parameters do not match data read
        from the device."""
        print("Running test_mismatched_channels")
        device = DsiDevice(connection_params=self.connection_params(),
                           channels=['ch1', 'ch2'])
        self.assertEqual(len(device.channels), 2)
        device.connect()

        with pytest.raises(Exception):
            device.acquisition_init()

    def test_mismatched_frequency(self):
        """An exception should be thrown if parameters do not match data read
        from the device."""
        print("Running test_mismatched_frequency")
        device = DsiDevice(connection_params=self.connection_params(), fs=100)
        self.assertEqual(device.fs, 100)
        device.connect()

        with pytest.raises(Exception):
            device.acquisition_init()

    def test_update_params_on_init(self):
        """Channel and sample rate properties should be updated by reading
        initialization data from the server."""
        print("Running test_update_params_on_init")
        device = DsiDevice(
            connection_params=self.connection_params(), channels=[])
        self.assertEqual(device.fs, dsi.DEFAULT_FS)
        self.assertEqual(len(device.channels), 0)

        device.connect()
        device.acquisition_init()

        self.assertEqual(device.fs, dsi.DEFAULT_FS)
        self.assertEqual(len(device.channels), len(dsi.DEFAULT_CHANNELS))

    def test_read_data(self):
        """Should produce a valid sensor_data record."""

        print("Running test_read_data")
        device = DsiDevice(connection_params=self.connection_params())

        device.connect()
        device.acquisition_init()
        data = device.read_data()

        self.assertTrue(len(data) > 0)
        self.assertEqual(len(data), len(device.channels))
        for channel in data:
            self.assertTrue(isinstance(channel, float))


if __name__ == '__main__':
    unittest.main()



import logging

REGISTERED_FEEDBACK_TYPES = ['sound', 'visual']


class Feedback:
    """Feedback."""
    def __init__(self, feedback_type):
        super(Feedback, self).__init__()
        self.feedback_type = feedback_type
        self.logger = logging.getLogger(__name__)

    def configure(self):
        raise NotImplementedError()

    def administer(self):
        raise NotImplementedError()

    def _type(self):
        return self.feedback_type

    def _available_modes(self):
        return REGISTERED_FEEDBACK_TYPES



import logging
import sys
import math
from typing import List
from collections import defaultdict
from bcipy.helpers.task import alphabet, SPACE_CHAR
from bcipy.language_model import lm_server
from bcipy.language_model.errors import (EvidenceDataStructError,
                                         NBestError,
                                         NBestHighValue)
from bcipy.language_model.lm_server import LmServerConfig
log = logging.getLogger(__name__)
sys.path.append('.')
ALPHABET = alphabet()
LM_SPACE = '#'

class LangModel:

    DEFAULT_CONFIG = LmServerConfig(
        image="oclmimage:version2.0",
        port=6000,
        docker_port=5000)

    def __init__(self, server_config: LmServerConfig=DEFAULT_CONFIG,
                 logfile: str = "log"):
        """
        Initiate the langModel class and starts the corresponding docker
        server for the given type.

        Input:
          lmtype - language model type
          logfile - a valid filename to function as a logger
        """
        self.server_config = server_config
        self.priors = defaultdict(list)
        
        log.setLevel(logging.INFO)
        log.addHandler(logging.FileHandler(logfile))

        lm_server.start(server_config)

    def init(self, nbest: int = 1):
        """
        Initialize the language model (on the server side)
        Input:
            nbest - top N symbols from evidence
        """
        if not isinstance(nbest, int):
            raise NBestError(nbest)
        if nbest > 4:
            raise NBestHighValue(nbest)
        lm_server.post_json_request(
            self.server_config, 'init', data={'nbest': nbest})

    def cleanup(self):
        """Stop the docker server"""
        lm_server.stop(self.server_config)

    def reset(self):
        """
        Clean observations of the language model use reset
        """
        lm_server.post_json_request(self.server_config, 'reset')
        self.priors = defaultdict(list)
        log.info("\ncleaning history\n")

    def state_update(self, evidence: List, return_mode: str = 'letter'):
        """
        Provide a prior distribution of the language model
        in return to the system's decision regarding the
        last observation
        Both lm types allow providing more the one timestep
        input. Pay attention to the data struct expected.
        OCLM
        Input:
            evidence - a list of (list of) tuples [[(sym1, prob), (sym2, prob2)]]
            the numbers are assumed to be in the log probabilty domain
            return_mode - 'letter' or 'word' (available
                          for oclm) strings
        Output:
            priors - a json dictionary with Normalized priors
                     in the Negative Log probabilty domain.
        """

        # assert the input contains a valid symbol
        try:
            clean_evidence = []
            for tmp_evidence in evidence:
                tmp = []
                for (symbol, pr) in tmp_evidence:
                    assert symbol in ALPHABET, \
                        "%r contains invalid symbol" % evidence
                    if symbol == SPACE_CHAR:
                        tmp.append((LM_SPACE, pr))
                    else:
                        tmp.append((symbol.lower(), pr))
                clean_evidence.append(tmp)
        except:
            raise EvidenceDataStructError

        output = lm_server.post_json_request(self.server_config,
                                             'state_update',
                                             {'evidence': clean_evidence,
                                              'return_mode': return_mode})
        return self.__return_priors(output, return_mode)

    def _logger(self):
        """
        Log the priors given the recent decision
        """
        # print a json dict of the priors
        log.info('\nThe priors are:\n')
        for k in self.priors.keys():
            priors = self.priors[k]
            log.info('\nThe priors for {0} type are:\n'.format(k))
            for (symbol, pr) in priors:
                log.info('{0} {1:.4f}'.format(symbol, pr))

    def recent_priors(self, return_mode='letter'):
        """
        Display the priors given the recent decision
        """

        if not bool(self.priors[return_mode]):
            output = lm_server.post_json_request(self.server_config,
                                                 'recent_priors',
                                                 {'return_mode': return_mode})
            return self.__return_priors(output, return_mode)
        else:
            return self.priors

    def __return_priors(self, output, return_mode):
        """
        A helper function to provide the desired output
        depending on the return_mode
        """

        self.priors = defaultdict(list)
        self.priors['letter'] = [
            (letter.upper(), prob)
            if letter != LM_SPACE
            else (SPACE_CHAR, prob)
            for (letter, prob) in output['letter']]

        if return_mode != 'letter':
            self.priors[return_mode] = list(map(tuple, output[return_mode]))
        return self.priors



import errno
import os
import shutil
import time
import unittest

from bcipy.helpers.save import init_save_data_structure


class TestSave(unittest.TestCase):
    """This is Test Case for Saving BCI data."""

    def setUp(self):
        # set up the needed paths and initial data save structure

        self.data_save_path = 'data/'
        self.user_information = 'test_user_002'
        self.parameters_used = 'bcipy/parameters/parameters.json'

        self.save_folder_name = init_save_data_structure(
            self.data_save_path,
            self.user_information,
            self.parameters_used)

    def tearDown(self):
        # clean up by removing the data folder we used for testing
        shutil.rmtree(self.data_save_path)

    def test_init_save_data_structure_creates_correct_save_folder(self):

        # assert the save folder was created
        self.assertTrue(os.path.isdir(self.save_folder_name))

    def test_parameter_file_copies(self):

        # construct the path of the parameters
        param_path = self.save_folder_name + "/parameters.json"

        # assert that the params file was created in the correct location
        self.assertTrue(os.path.isfile(param_path))

    def test_throws_useful_error_if_given_incorrect_params_path(self):

        # try passing a parameters file that does not exist
        with self.assertRaises(Exception):
            init_save_data_structure(
                self.data_save_path,
                'new_user',
                'does_not_exist.json')

    def test_init_save_data_structure_makes_helpers_folder(self):

        # contruct the path of the helper folder
        helper_folder_name = self.save_folder_name + '/helpers/'

        # attempt to make that folder
        try:
            os.makedirs(helper_folder_name)

        except OSError as error:
            # assert the error returned, is that the dir exists.
            self.assertEqual(error.errno, errno.EEXIST)

if __name__ == '__main__':
    unittest.main()


"""Tests for Datastream Producer"""
import queue
import time

import unittest
from bcipy.acquisition.datastream.producer import Producer


class TestProducer(unittest.TestCase):
    """Tests for Producer"""

    def test_frequency(self):
        """Data should be generated at the provided frequency"""
        sample_hz = 300
        runtime = 0.2
        data_queue = queue.Queue()
        producer = Producer(data_queue, freq=1 / sample_hz)
        producer.start()
        time.sleep(runtime)
        producer.stop()

        data_n = data_queue.qsize()
        expected_n = sample_hz * runtime
        tolerance = 5
        self.assertTrue(data_n + tolerance >= expected_n)
        self.assertTrue(data_n <= expected_n + tolerance)

    def test_custom_generator(self):
        """Producer should be able to take a custom generator."""

        def gen():
            counter = 0
            while True:
                counter += 1
                yield counter

        data_queue = queue.Queue()
        producer = Producer(data_queue, freq=1 / 300, generator=gen())
        producer.start()
        time.sleep(0.1)
        producer.stop()

        lst = list(data_queue.queue)
        self.assertTrue(len(lst) > 0)
        self.assertEqual(lst[0], 1)
        self.assertEqual(lst[-1], len(lst))
        print(lst[-1])

    def test_max_iters(self):
        """Producer should stop producing data after maxiters if param is
        provided."""

        sample_hz = 300
        runtime = 0.2
        maxiters = 10
        data_queue = queue.Queue()
        producer = Producer(data_queue, freq=1 / sample_hz, maxiters=maxiters)
        producer.start()
        time.sleep(runtime)
        producer.stop()

        expected_n = sample_hz * runtime
        tolerance = 10
        self.assertTrue(expected_n - tolerance > maxiters)
        data_n = data_queue.qsize()
        self.assertEqual(data_n, maxiters)
